#include "BulletTypes.h"
#include "Effect.h"
#include "Crawler.h"
#include "DEFINES.h"
#include "Emitter.h"
#include "utils.h"

INCLUDE_game
INCLUDE_MONSTER_LIST
INCLUDE_EMITTER_LIST

LightningBolt::LightningBolt(vf2d pos,vf2d vel,float radius,int damage,bool upperLevel,bool friendly,Pixel col)
	:Bullet(pos,vel,radius,damage,AnimationState::LIGHTNING_BOLT,upperLevel,false,INFINITE,true,friendly,col){}

void LightningBolt::Update(float fElapsedTime){
	lastParticleSpawn=std::max(0.f,lastParticleSpawn-fElapsedTime);
	if(lastParticleSpawn==0){
		lastParticleSpawn=0.01;
		uint8_t brightness=uint8_t(util::random(100)+150);
		switch(rand()%4){
			case 0:{
				game->AddEffect(std::make_unique<Effect>(pos+vf2d{util::random(12)-6,util::random(12)-6},util::random(0.1),AnimationState::LIGHTNING_BOLT_PARTICLE1,upperLevel,util::random(0.5)+1,0.1,vel*(util::random(0.1)+0.9),Pixel{brightness,brightness,brightness}));
			}break;
			case 1:{
				game->AddEffect(std::make_unique<Effect>(pos+vf2d{util::random(12)-6,util::random(12)-6},util::random(0.1),AnimationState::LIGHTNING_BOLT_PARTICLE2,upperLevel,util::random(0.5)+1,0.1,vel*(util::random(0.1)+0.9),Pixel{brightness,brightness,brightness}));
			}break;
			case 2:{
				game->AddEffect(std::make_unique<Effect>(pos+vf2d{util::random(12)-6,util::random(12)-6},util::random(0.1),AnimationState::LIGHTNING_BOLT_PARTICLE3,upperLevel,util::random(0.5)+1,0.1,vel*(util::random(0.1)+0.9),Pixel{brightness,brightness,brightness}));
			}break;
			case 3:{
				game->AddEffect(std::make_unique<Effect>(pos+vf2d{util::random(12)-6,util::random(12)-6},util::random(0.1),AnimationState::LIGHTNING_BOLT_PARTICLE4,upperLevel,util::random(0.5)+1,0.1,vel*(util::random(0.1)+0.9),Pixel{brightness,brightness,brightness}));
			}break;
		}
	}
}

bool LightningBolt::PlayerHit(Player*player)
{
	deactivated=true;
	fadeOutTime=0.2f;
	game->AddEffect(std::make_unique<Effect>(player->GetPos(),0.3,AnimationState::LIGHTNING_SPLASH,upperLevel,player->GetSizeMult(),0.25,vf2d{},WHITE,util::random(PI)));
	return false;
}

bool LightningBolt::MonsterHit(Monster& monster)
{
	deactivated=true;
	fadeOutTime=0.2f;
	game->AddEffect(std::make_unique<Effect>(monster.GetPos(),0.3,AnimationState::LIGHTNING_SPLASH,upperLevel,monster.GetSizeMult(),0.25,vf2d{},WHITE,util::random(PI)));
	int targetsHit=0;
	for(Monster&m:MONSTER_LIST){
		if(&m==&monster||monster.OnUpperLevel()!=m.OnUpperLevel())continue;
		geom2d::line<float>lineToTarget=geom2d::line<float>(monster.GetPos(),m.GetPos());
		float dist=lineToTarget.length();
		vf2d vec;
		vec.norm();
		if(dist<=72){
			if(m.Hurt(game->GetPlayer()->GetAttack()*2)){
				EMITTER_LIST.push_back(std::make_unique<LightningBoltEmitter>(LightningBoltEmitter(monster.GetPos(),m.GetPos(),0.05,0.25,upperLevel)));
				game->AddEffect(std::make_unique<Effect>(m.GetPos(),0.5,AnimationState::LIGHTNING_SPLASH,upperLevel,monster.GetSizeMult(),0.25,vf2d{},WHITE,util::random(PI)));
				targetsHit++;
			}
		}
		if(targetsHit>=2)break;
	}
	return false;
}